"""
XPRESStools
A toolkit for navigating and analyzing gene expression datasets
alias: xpresstools

Copyright (C) 2019  Jordan A. Berg
jordan <dot> berg <at> biochem <dot> utah <dot> edu

This program is free software: you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

This program is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE. See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along with
this program.  If not, see <https://www.gnu.org/licenses/>.
"""

"""
IMPORT DEPENDENCIES
"""
import pandas as pd
import numpy as np
import scipy.stats as stats
import matplotlib
import matplotlib.pyplot as plt
import matplotlib.patches as patches
from mpl_toolkits.mplot3d import Axes3D
import seaborn as sns
from functools import partial
from multiprocessing import cpu_count, Pool

"""
DESCRIPTION: Default data prep for all analysis functions
"""
def analysis_prep(data):
    data_c = data.copy()
    data_c = data_c.dropna(axis=0)
    return data_c

"""
DESCRIPTION: Axis-agnostic list-reader from dataframe
VARIABLES:
USAGE:
ASSUMPTIONS:
"""
def custom_list(file, delimiter=','):

    gene_df = pd.read_csv(str(file), sep=delimiter, index_col=False, header=None)
    genes = gene_df.values.tolist()
    gene_list = [val for sublist in genes for val in sublist]

    return gene_list

"""
DESCRIPTION: Reset plotting object to avoid bleed through
"""
def reset_plot(whitegrid, ax=False):

    if ax == True:
        del ax

    plt.close()
    plt.clf()

    if whitegrid == True:
        sns.set_style("whitegrid")
    else:
        sns.set_style("darkgrid")

"""
DESCRIPTION: Return the subsetted dataframe using a gene list
"""
def data_subset(data, gene_list):

    data_c = data.copy()

    #Check file formats
    if type(gene_list) is list:
        data_sub = data_c.reindex(labels=gene_list, axis=0)
    elif type(gene_list) is str:
        genes = custom_list(str(gene_list))
        data_sub = data_c.reindex(labels=genes, axis=0)
    else:
        print('Incorrect gene_list type provided')
        return

    return data_sub

"""
DESCRIPTION: Prepare a color map from palette dictionary
"""
def prep_palette(info, palette):

    if type(palette) is dict:
        info = info.T
        info.columns = info.iloc[0]
        info = info.reindex(info.index.drop(0))
        info = info.rename({1: 'samples'})
        labels = info.iloc[0]
        color_map = labels.map(palette)
        return color_map
    else:
        print('Error: a dictionary was not provided as palette')
        return

"""
DESCRIPTION: Add labels to dataframe for each sample
"""
def label(data, info):

    data_c = data.copy()

    #Prep data_scaled by adding labels from info
    labels = pd.Series(info[1].values,index=info[0]).to_dict()
    data_c.loc['label'] = data_c.columns.map(labels.get)

    #Output collapsed dataframe
    newIndex = ['label'] + [ind for ind in data_c.index if ind != 'label']
    data_c = data_c.reindex(index=newIndex)
    data_c = data_c.T.set_index('label', drop=True)
    data_c = data_c.T.dropna(axis=0)

    return data_c

"""
DESCRIPTION: Unstack dataframe for use with some seaborns functions
"""
def unstack_data(data):

    data_c = data.copy()
    data_unstacked = data_c.unstack().reset_index()
    data_unstacked = data_unstacked.rename(columns={data_unstacked.columns[0]: 'type', data_unstacked.columns[1]: 'gene', data_unstacked.columns[2]: 'expr'})

    return data_unstacked

"""
DESCRIPTION: Convert gene in dataframe to array
"""
def get_array(data, gene):

    data_c = data.copy()

    gene = data_c.loc[str(gene)].values.tolist()
    gene = np.array(gene).astype(np.float)
    gene = np.ndarray.tolist(gene)

    return gene

"""
"""
def make_linreg(data, gene1, gene2):

    data_c = data.copy()
    gene_a = get_array(data_c, gene1)
    gene_b = get_array(data_c, gene2)

    slope, intercept, r_value, p_value, std_err = stats.linregress(gene_a, gene_b)
    x = np.linspace(data_c.loc[str(gene1)].min(), data_c.loc[str(gene1)].max(), 100)
    y = (slope * x) + intercept

    title = 'r = ' + "%.2f" % round(r_value,4)

    return x, y, r_value, title

"""
"""
def add_x_threshold(threshold, threshold_color, ax):

    for x in threshold:
        if type(x) is int or type(x) is float:
            ax.axvline(x, ls='--', color=str(threshold_color))
        else:
            print('Invalid X threshold provided')

def add_y_threshold(threshold, threshold_color, ax):

    for x in threshold:
        if type(x) is int or type(x) is float:
            ax.axhline(x, ls='--', color=str(threshold_color))
        else:
            print('Invalid Y threshold provided')

"""
Initial variable checks
"""
def init_pca(principle_components, _3d_pca, plotly_login):

    if len(principle_components) != 2 and _3d_pca == False:
        print('Incompatible options provided for principle_components and _3d_pca')
        return

    if plotly_login != None and _3d_pca == False:
        print('Only provide plotly login for _3d_pca')
        return

    if _3d_pca == True:
        if principle_components == [1,2]:
            principle_components = [1,2,3]

        elif len(principle_components) != 3:
            print('Incompatible options provided for principle_components and _3d_pca')
            return

        else:
            print('Not sure if this will ever be triggered, but if it is need to check case')
            return

        if plotly_login != None and type(plotly_login) is not list:
            print('Provide proper plotly login information in form of list')
            print("['userid','api key']")
            return

    return principle_components

"""
DESCRIPTION: Generate scree plot for principle components
"""
def make_scree(pca, n_components, save_fig, dpi, bbox_inches, scree_only, save_scree, grid, whitegrid):

    vari = 'Explained variation per principal component: {}'.format(np.round(pca.explained_variance_ratio_, decimals=4)*100)
    scree = np.round(pca.explained_variance_ratio_, decimals=4)*100

    #Plot scree
    sing_vals = np.arange(n_components) + 1

    ax = sns.lineplot(x=sing_vals, y=scree, color="red")
    ax.set(xlabel='Principal Component', ylabel='Proportion of Variance Explained', title='Scree Plot')

    if save_scree == True:
        plt.savefig(str(save_fig[:-4]) + '_scree.png', dpi=dpi, bbox_inches=bbox_inches)

    if scree_only == True:
        plt.show()

    if grid == False:
        ax.grid(False)

    #Remove scree from memory to prevent plot bleeding
    reset_plot(whitegrid, ax=True)

    return scree

"""
DESCRIPTION: Add confidence intervals to scatterplot
NOTE: This code is adapted from Jaime (https://stackoverflow.com/a/20127387/9571488) and Ben (https://stackoverflow.com/a/25022642/9571488) on Stack Overflow
"""
def set_confidence(df_pca, pca_plot, unique_labels, palette, ci):

    for x in unique_labels:
        #slice df into label specific datasets
        df_slice = df_pca[df_pca['label'] == x]

        #make numpy array from label-specific dfs
        x_slice = df_slice.PCa.values
        y_slice = df_slice.PCb.values

        #pca maths
        cov = np.cov(x_slice, y_slice)
        lambda_, v = np.linalg.eig(cov)
        theta = np.degrees(np.arctan2(*v[:,0][::-1]))
        lambda_ = np.sqrt(lambda_)

        #plot
        pca_plot.add_patch(patches.Ellipse(xy=(np.mean(x_slice), np.mean(y_slice)),
                          width=lambda_[0]*ci*2, height=lambda_[1]*ci*2,
                          angle=theta,
                          alpha=0.3, facecolor=palette[x], edgecolor='black', linewidth=1, linestyle='solid')
                          )

"""
DESCRIPTION: 2D Non-interactive PCA scatterplot
"""
def pca2(df_pca, unique_labels, palette, principle_components, scree, order_legend, save_fig, dpi, bbox_inches, ci, grid, title, size, highlight_color, highlight_names):

    ax = sns.scatterplot(df_pca.PCa, df_pca.PCb, hue=df_pca['label'], palette=palette, s=size)
    set_confidence(df_pca, ax, unique_labels, palette, ci)

    # Put the legend out of the figure
    make_legend(ax, order_legend, highlight_color, highlight_names)

    plt.xlabel('PC' + str(principle_components[0]) + ' (' + str(round(scree[(principle_components[0] - 1)],2)) + '%)')
    plt.ylabel('PC' + str(principle_components[1]) + ' (' + str(round(scree[(principle_components[1] - 1)],2)) + '%)')

    if grid == False:
        plt.grid(False)

    if save_fig != None:
        #Save plot
        plt.title(str(title))
        plt.savefig(str(save_fig), dpi=dpi, bbox_inches=bbox_inches)

    plt.show()

"""
DESCRIPTION: 3D Non-interactive PCA scatterplot
"""
def pca3(df_pca, palette, save_fig, dpi, bbox_inches, size, order_legend):

    fig = plt.figure()
    ax = Axes3D(fig)

    pc_list = df_pca.columns.tolist()
    while 'label' in pc_list:
        pc_list.remove('label')

    df_pca.columns = ['PCa', 'PCb', 'PCc', 'label']
    unique_labels = df_pca['label'].unique() #Gather unique labels

    #Convert color tuples to arrays

    for x in unique_labels:
        pca0 = df_pca.loc[df_pca['label'] == str(x)]
        x0 = pca0.PCa.values
        y0 = pca0.PCb.values
        z0 = pca0.PCc.values
        ax.scatter(x0, y0, z0, c=np.array([palette[str(x)]]), label=str(x), s=size)

    ax.set_xlabel(str(pc_list[0]))
    ax.set_ylabel(str(pc_list[1]))
    ax.set_zlabel(str(pc_list[2]))

    plt.show()

"""
DESCRIPTION: Parallelize function on a chunk of a dataframe
"""
def parallelize(func, *args):

    cores = cpu_count() #Number of CPU cores on your system
    partitions = cpu_count() #Define as many partitions as you want

    data_split = np.array_split(args[0], partitions)
    pool = Pool(cores)

    if func == calculate_fc:
        func = partial(calculate_fc, label_comp=args[1], label_base=args[2])
    elif func == calculate_p:
        func = partial(calculate_p, label_comp=args[1], label_base=args[2])
    elif func == count_threshold_util:
        func = partial(count_threshold_util, minimum=args[1], maximum=args[2])
    else:
        return

    data = pd.concat(pool.map(func, data_split))

    pool.close()
    pool.join()

    return data

"""
DESCRIPTION
"""
def calculate_fc(data, label_comp, label_base):

    # Average every by cell line
    data['log$_2$(Fold Change)'] = np.log2((data.filter(regex=str(label_comp)).mean(axis=1)) / \
                                      (data.filter(regex=str(label_base)).mean(axis=1)))
    data['-log$_1$$_0$(P-Value)'] = ''

    return data

"""
DESCRIPTION
"""
def calculate_p(data, label_comp, label_base):

    drop_index = []

    # Calculate p-value using 1-way ANOVA with replicates and append to df_oxsm_volc
    for row in data.iterrows():
        index, row_data = row
        comp_row = data.loc[index].filter(regex=str(label_comp)).values.tolist()
        base_row = data.loc[index].filter(regex=str(label_base)).values.tolist()

        if len(comp_row) < 2 or len(base_row) < 2:
            raise Exception('Calculating a P-value requires replicates for each sample type')

        # Append p_value to df_oxsm_volc
        try:
            statistic, p_value = stats.ttest_ind(comp_row, base_row)
            data.loc[index,'-log$_1$$_0$(P-Value)'] = float(-1 * (np.log10(p_value)))
        except:
            drop_index.append(index)

    data = data.drop(labels=drop_index, axis=0)

    return data

"""
DESCRIPTION
"""
def count_threshold_util(data, minimum, maximum):

    data = data.T

    if minimum != None:
        data = data[data.columns[data.min() > minimum]]

    if maximum != None:
        data = data[data.columns[data.max() < maximum]]

    data = data.T

    return data

"""
"""
def make_threshold_list(threshold):

    if threshold != None and type(threshold) != list:
        threshold = [threshold]

    return threshold

"""

"""
def output_threshold(data_c, x_threshold, y_threshold, save_threshold_hits, save_threshold_hits_delimiter):

    x_threshold = make_threshold_list(x_threshold)
    y_threshold = make_threshold_list(y_threshold)

    if len(x_threshold) == 2 and len(y_threshold) == 1:
        data_c = data_c.rename(columns = {'log$_2$(Fold Change)':'log2 Fold Change', '-log$_1$$_0$(P-Value)':'-log10 P-Value'})
        df_c = data_c[['log2 Fold Change', '-log10 P-Value']].copy()
        df_up = df_c.loc[(df_c['log2 Fold Change'] > max(x_threshold)) & (df_c['-log10 P-Value'] > y_threshold[0])] #get upregulated hits
        df_down = df_c.loc[(df_c['log2 Fold Change'] < min(x_threshold)) & (df_c['-log10 P-Value'] > y_threshold[0])] #get downregulated hits
        thresh_hits = df_up.append(df_down) #append hits tables
        thresh_hits.to_csv(str(save_threshold_hits), sep=save_threshold_hits_delimiter)

"""
"""
def make_legend(ax, order_legend, highlight_color, highlight_names):

    handles, labels = ax.get_legend_handles_labels()

    if len(labels) > 1:
        if order_legend != None:
            if type(order_legend) is list:
                plt.legend([handles[idx] for idx in order_legend],[labels[idx] for idx in order_legend], bbox_to_anchor=(1.05, 1), loc=2, borderaxespad=0.)
            else:
                plt.legend(handles, labels, bbox_to_anchor=(1.05, 1), loc=2, borderaxespad=0.)
                print('order_legend datatype is invalid -- plotting samples in default order...')
        else:
            plt.legend(handles, labels, bbox_to_anchor=(1.05, 1), loc=2, borderaxespad=0.)

    elif len(labels) == 0:
        if highlight_names != None and type(highlight_names) is list:
            if type(highlight_color) is str:
                highlight_color = [highlight_color]

            if len(highlight_names) == len(highlight_color):
                #Make a custom legend using the provided info
                #Adapted from https://stackoverflow.com/a/47749903
                f = lambda m,c: plt.plot([],[],marker='o', color=c, ls="none")[0]
                handles = [f("s", highlight_color[i]) for i in range(len(highlight_color))]
                plt.legend(handles, highlight_names, bbox_to_anchor=(1.05, 1), loc=2, borderaxespad=0.)
            else:
                raise Exception('Highlight colors and names lists are not of equal size')
        else:
            pass #raise Exception('Highlight names is not a list')

    else:
        pass #ax.get_legend().remove()

    return ax

"""
"""
def highlight_markers(data, x, y, highlight_points, highlight_color, highlight_names, alpha_highlights, ax):

    if all(isinstance(z, list) for z in highlight_points):
        p = 0
        for h in highlight_points:
            data_sub = data_subset(data, highlight_points[p]).T
            data_genes = data_sub.dropna(axis=1)
            if type(alpha_highlights) is list:
                ax = sns.scatterplot(x=data_genes.loc[str(x)], y=data_genes.loc[str(y)], color=str(highlight_color[p]), alpha=alpha_highlights[p])
            else:
                ax = sns.scatterplot(x=data_genes.loc[str(x)], y=data_genes.loc[str(y)], color=str(highlight_color[p]), alpha=alpha_highlights)
            p += 1
    else:
        data_sub = data_subset(data, highlight_points).T
        data_genes = data_sub.dropna(axis=1)
        ax = sns.scatterplot(x=data_genes.loc[str(x)], y=data_genes.loc[str(y)], color=str(highlight_color), alpha=alpha_highlights)

    return ax



"""
XPRESStools
A toolkit for navigating and analyzing gene expression datasets
alias: xpresstools

Copyright (C) 2019  Jordan A. Berg
jordan <dot> berg <at> biochem <dot> utah <dot> edu

This program is free software: you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

This program is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE. See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along with
this program.  If not, see <https://www.gnu.org/licenses/>.
"""

from .fetch import *
from .format import *
from .normalize import *
from .collapser import *
from .analyze import *
from .collapser import *
from .interactive import *
from .utils_analyze import *
from .utils_truncator import *
from .utils import *


__version__ = '0.0.3-beta'



"""
XPRESStools
A toolkit for navigating and analyzing gene expression datasets
alias: xpresstools

Copyright (C) 2019  Jordan A. Berg
jordan <dot> berg <at> biochem <dot> utah <dot> edu

This program is free software: you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

This program is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE. See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along with
this program.  If not, see <https://www.gnu.org/licenses/>.
"""

"""
IMPORT DEPENDENCIES
"""
import plotly
import plotly.plotly as py
import plotly.graph_objs as go

"""
DESCRIPTION: Create 3D scatterplot for PCA with plotly
"""
def interactive_3D(df_pca, size, palette, plotly_login, save_fig):

    plotly.tools.set_credentials_file(username=str(plotly_login[0]), api_key=str(plotly_login[1]))

    df_pca.columns = ['PCa', 'PCb', 'PCc', 'label']
    unique_labels = df_pca['label'].unique() #Gather unique labels

    pca0 = df_pca.loc[df_pca['label'] == unique_labels[0]]
    pca1 = df_pca.loc[df_pca['label'] == unique_labels[1]]
    pca2 = df_pca.loc[df_pca['label'] == unique_labels[2]]

    x0 = pca0.PCa.values
    y0 = pca0.PCb.values
    z0 = pca0.PCc.values
    trace0 = go.Scatter3d(
        x=x0,
        y=y0,
        z=z0,
        name=unique_labels[0],
        mode='markers',
        marker=dict(
            size=size,
            color=palette[unique_labels[0]],
            opacity=0.8
        )
    )

    x1 = pca1.PCa.values
    y1 = pca1.PCb.values
    z1 = pca1.PCc.values
    trace1 = go.Scatter3d(
        x=x1,
        y=y1,
        z=z1,
        name=unique_labels[1],
        mode='markers',
        marker=dict(
            size=size,
            color=palette[unique_labels[1]],
            opacity=0.8
        )
    )

    x2 = pca2.PCa.values
    y2 = pca2.PCb.values
    z2 = pca2.PCc.values
    trace2 = go.Scatter3d(
        x=x2,
        y=y2,
        z=z2,
        name=unique_labels[2],
        mode='markers',
        marker=dict(
            size=size,
            color=palette[unique_labels[2]],
            opacity=0.8
        )
    )

    data = [trace0, trace1, trace2]
    layout = go.Layout(
        margin=dict(
            l=0,
            r=0,
            b=0,
            t=0
        ),
        hovermode='closest'
    )
    fig = go.Figure(data=data, layout=layout)

    if save_fig != None:
        py.offline.plot(fig, filename=str(save_fig))
    else:
        py.iplot(fig, filename='3d_pca')

"""
DESCRIPTION: Create interactive scatter plot that displays relevant sample/gene name and coordinates

VARIABLES:
data= MICARtools formatted expression matrix
"""
def interactive_scatter(data, info, x, y, plotly_login, palette, highlight='sample', save_fig=None):

    plotly.tools.set_credentials_file(username=plotly_login[0], api_key=plotly_login[1])
    data_c = data.copy()

    if highlight == 'sample':
        #Add labels if not there
        if 'label' not in data_c.index:
            labels = pd.Series(info[1].values,index=info[0]).to_dict()
            data_c.loc['label'] = data_c.columns.map(labels.get)

        data_c = data_c.T
        df_plot = data_c[[x, y,'label']]

        traces = []
        for index, row in df_plot.iterrows():

            _x = [df_plot.loc[index][x]]
            _y = [df_plot.loc[index][y]]

            traces.append(go.Scatter(
                                x = _x,
                                y = _y,
                                name = index,
                                mode = 'markers',
                                marker = dict(
                                    size = 10,
                                    color = palette[df_plot.loc[index]['label']],
                                    line = dict(
                                        width = 2,
                                        color = 'rgb(0, 0, 0)'
                                    )
                                )
                            )
                         )

    elif highlight == 'gene':
        df_plot = data_c[[x, y]]

        traces = []
        for index, row in df_plot.iterrows():

            _x = [df_plot.loc[index][x]]
            _y = [df_plot.loc[index][y]]

            traces.append(go.Scatter(
                                x = _x,
                                y = _y,
                                name = index,
                                mode = 'markers',
                                marker = dict(
                                    size = 2,
                                    line = dict(
                                        width = 2
                                    )
                                )
                            )
                         )

    else:
        print('Error: Invalid input for highlight')
        return

    data = traces
    layout = dict(yaxis = dict(zeroline = False),
                  xaxis = dict(zeroline = False),
                  showlegend=False,
                  hovermode='closest'
                 )

    fig = dict(data=data, layout=layout)
    if save_fig != None:
        py.offline.plot(fig, filename=str(save_fig))
    else:
        py.iplot(fig, filename='2d_scatter')



"""
XPRESStools
A toolkit for navigating and analyzing gene expression datasets
alias: xpresstools

Copyright (C) 2019  Jordan A. Berg
jordan <dot> berg <at> biochem <dot> utah <dot> edu

This program is free software: you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

This program is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE. See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along with
this program.  If not, see <https://www.gnu.org/licenses/>.
"""

"""
IMPORT DEPENDENCIES
"""
import csv
import pandas as pd
from .utils import check_directories
from .utils_truncator import execute_truncator, parallelize_truncator
pd.options.mode.chained_assignment = None

"""
DESCRIPTION: Convert row names (genes) of dataframe using GTF as reference for new name

RETURNS: Return the renamed dataframe

VARIABLES:
data= Dataframe to convert rows names
gtf= Path and name of gtf reference file
orig_name_label= Label of original name (usually a \"gene_id \"')
orig_name_location= Position in last column of GTF where relevant data is found (i.e. 0 would be the first sub-string before the first comma, 3 would be the third sub-string after the second comma before the third comma)
new_name_label= Label of original name (usually \"gene_name \")
new_name_location= Position in last column of GTF where relevant data is found (i.e. 0 would be the first sub-string before the first comma, 3 would be the third sub-string after the second comma before the third comma)
refill= In some cases, where common gene names are unavailable, the dataframe will fill the gene name with the improper field of the GTF. In this case, specify this improper string and these values will be replaced with the original name
sep= GTF delimiter (usually tab-delimited)
"""
def convert_names_gtf(data, gtf, orig_name_label='gene_id \"', orig_name_location=0, new_name_label='gene_name \"', new_name_location=1, refill=None, sep='\t'):

    #Import reference GTF
    gtf = pd.read_csv(str(gtf),sep=sep,comment='#', low_memory=False, header=None)

    #Parse out old and new names from GTF
    gtf_genes = gtf.loc[gtf[2] == 'gene']
    gtf_genes['original'] = gtf[8].str.split(';').str[orig_name_location]
    gtf_genes['new'] = gtf[8].str.split(';').str[new_name_location]
    gtf_genes['original'] = gtf_genes['original'].map(lambda x: x.lstrip(str(orig_name_label)).rstrip('\"').rstrip(' '))
    gtf_genes['new'] = gtf_genes['new'].map(lambda x: x.lstrip(str(new_name_label)).rstrip('\"').rstrip(' '))
    gtf_genes = gtf_genes[['original','new']].copy()

    #Create dictionary
    if refill != None:
        gene_dict = {}
        for index, row in gtf_genes.iterrows():
            if row[1] == str(refill):
                gene_dict[row[0]] = row[0]
            else:
                gene_dict[row[0]] = row[1]
    else:
        gene_dict = pd.Series(gtf_genes['new'].values,index=gtf_genes['original']).to_dict()

    #Replace old gene names/ids with new
    data_names = data.copy()
    data_names['new'] = data_names.index.to_series().map(gene_dict).fillna(data_names.index.to_series())
    data_names = data_names.set_index('new')
    del data_names.index.name

    return data_names



"""
XPRESStools
A toolkit for navigating and analyzing gene expression datasets
alias: xpresstools

Copyright (C) 2019  Jordan A. Berg
jordan <dot> berg <at> biochem <dot> utah <dot> edu

This program is free software: you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

This program is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE. See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along with
this program.  If not, see <https://www.gnu.org/licenses/>.
"""

"""
IMPORT DEPENDENCIES
"""
import pandas as pd
import numpy as np

"""
DESCRIPTION: Create dictionary of probes and their gene names for the microarray probe collapser

VARIABLES:
reference= Full path and file name for GPL reference file, accessed from NCBI (should be a .txt file)
gene_list= Full path and file name to .csv file listing gene names to get probes for. Resulting function dictionary will only contain these genes and their probes (default: None)
no_multimappers= Do not allow ambiguous probes in the probe-gene dictionary (default: True)

USAGE:
import xpresstools as xp
gene_dict = xp.prep_collapser("~/Desktop/GPL570.txt")

ASSUMPTIONS:
If using gene_list option, file must be a .csv
Assumes GPL .txt file from NCBI is tab delimited
"""
def prep_collapser(reference, gene_list=None, no_multimappers=True):

    #Get probe/gene_name reference file and import into a dataframe
    df = pd.read_csv(str(reference), sep="\t", low_memory=False, comment='#') #no index
    df = df[['ID','Gene Symbol']]
    df = df.dropna()

    #If a custom gene list is given so that only certain probes are looked at and collapsed downstream
    #Take the gene list file, join genes for regex search
    #Create dictionary for just the probes corresponding with the genes of interest
    if gene_list != None:
        if type(gene_list) == list:

            gene_list = custom_list(gene_list)

            search = '|'.join(gene_list)
            search = search.upper()

            df = df[df['Gene Symbol'].str.contains(search)] #only df where probes of interest are
            df_dict = df.set_index('ID')['Gene Symbol'].to_dict()
    #Take full probe set and create dictionary
    else:
        df_dict = df.set_index('ID')['Gene Symbol'].to_dict()

    #If allowing for multimappers, return dictionary as is
    if no_multimappers == False:
        return df_dict
    #Remove any probes that map to several genes
    else:
        df_dict_mod = {}
        for key, value in df_dict.items():
            if '///' in value:
                continue
            else:
                df_dict_mod[key] = value

        return df_dict_mod

"""
DESCRIPTION: Collapse probes of microarray dataset using previously prepared probe collapser dictionary (see prep_collapser function))

METHODS: Works by taking all probe expression data of a particular gene and determines a mean expression value for each sample for the given gene

VARIABLES:
data= Dataframe of microarray probe data to be collapsed
dict= Probe collapser dictionary created in the prep_collapser function

USAGE:
import xpresstools as xp
df_collapsed = xp.probe_collapse(df, collapser_dict)

ASSUMPTIONS:
A probe collapser dictionary has been previously prepared using the prep_collapser function
"""
def run_collapser(data, dict):

    #This appears to fix the SettingwithCopyWarning error
    data_c = data.copy()

    #Get list of probes to find in df
    dict_list = list(dict.keys())

    #only keep df rows where probes of interest are
    #(in cases where only looking at certain genes, default: all genes)
    #column 'name' header for probes in these files
    try:
        data_c = data_c[data_c['name'].isin(dict_list)]
    except:
        data_c['name'] = data_c.index
        data_c = data_c[data_c['name'].isin(dict_list)]

    #Map gene names in place of probes
    #Will set off SettingwithCopyWarning -- should be fine as we are replacing values in same column and it appears to change as expected
    data_c['name'] = data_c['name'].map(dict)

    #Set gene names as indices (allows for multiple indices with same name)
    #Needed to remove strings from df to allow for next step
    data_c = data_c.set_index('name', drop=True)

    #force data to float
    data_numeric = data_c.apply(pd.to_numeric)
    #Reset indices to its own column to allow for sorting in next step
    data_numeric['name_sort'] =  data_numeric.index

    #groupby index (gene name) and collapse, taking the mean of rows with same name in 'name_sort'
    #Sets name_sort column names (post-collapse) as indices
    data_collapsed = data_numeric.groupby('name_sort').mean()

    #Remove double header for indices
    del data_collapsed.index.name

    return data_collapsed

"""
DESCRIPTION: Ties prep_collapser and probe_collapse functions together

VARIABLES:
data= Dataframe of microarray probe data to be collapsed
reference= Full path and file name for GPL reference file, accessed from NCBI (should be a .txt file)
gene_list= Full path and file name to .csv file listing gene names to get probes for. Resulting function dictionary will only contain these genes and their probes (default: None)
no_multimappers= Do not allow ambiguous probes in the probe-gene dictionary (default: True)

USAGE:
import xpresstools as xp
df_collapsed = xp.auto_collapse(df, "~/Desktop/GPL570.csv")

ASSUMPTIONS:
See assumptions for prep_collapser and probe_collapse functions
"""
def probe_collapse(data, reference, gene_list=None, no_multimappers=True):

    dict = prep_collapser(reference, gene_list=gene_list, no_multimappers=no_multimappers)
    data_collapsed = run_collapser(data, dict)

    return data_collapsed



"""
XPRESStools
A toolkit for navigating and analyzing gene expression datasets
alias: xpresstools

Copyright (C) 2019  Jordan A. Berg
jordan <dot> berg <at> biochem <dot> utah <dot> edu

This program is free software: you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

This program is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE. See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along with
this program.  If not, see <https://www.gnu.org/licenses/>.
"""

"""
IMPORT DEPENDENCIES
"""
import os, sys
import re
import pandas as pd
import GEOparse
from .normalize import clean_df
from .utils import check_directories

"""
DESCRIPTION: Get dataframe from user file
VARIABLES:
file_name= full path of file to import into pandas dataframe
delimiter= delimiter type for importing file, default: ','
low_memory= Specify memory limits for importing large files, default: False (allows for large imports)
gene_axis= Orientiation of the data, where categorical data is either column-wise, (default: 'col') or row-wise ('row'). Case insensitive
USAGE:
import xpresstools as xp
data = xp.get_df("~/Desktop/data.csv")
ASSUMPTIONS:
Dataset does not contain axis labels (i.e. a column header for 'gene names')
Dataset only has gene names and sample_ids as column headers and row indices. Orientation is flexible, but needs to be specified in options if genes are not rows
If orientation is not default, it is then specified or else function will not be able to properly format the dataframe for downstream application
"""
def get_df(file_name, delimiter=",", low_memory=False, gene_axis='row'):

    #Read in file
    data = pd.read_csv(str(file_name), sep=delimiter, index_col=0, header=0, low_memory=low_memory)

    #Check data orientation
    if str(gene_axis).lower() == 'row':
        data = data
    elif str(gene_axis).lower() == 'col':
        data = data.T
    else:
        print("Incorrect gene_axis option specified")

    data = clean_df(data)

    return data

"""
DESCRIPTION: Get GEO dataframe and metadata
VARIABLES:
geo_id= GEO ID for dataset of interest, input is case insensitive (ex: GSE20716)
output_info= Output long-form metadata to txt file
USAGE:
sample_data, sample_metadata = xp.get_geo("GSE20716")

Add feature to grab GPL file
"""
def get_geo(geo_id, output_info=False, output_path="./"):

    #Get data
    gse = GEOparse.get_GEO(geo=str(geo_id).upper(), destdir=output_path) #Import GSE dataset

    data = gse.pivot_samples('VALUE')
    data = clean_df(data)

    #Get metadata
    #Write data to output file
    if output_info != False:
        with open(str(geo_id).upper() + '.txt', 'w+') as f: #Save all information as text file for reference
            for gsm_name, gsm in gse.gsms.items():
                f.write(gsm_name + '\n')
                for key, value in gsm.metadata.items():
                    f.write(" - %s : %s" % (key, ", ".join(value)) + '\n')

    #Populate metadata with sample ids and names
    metadata = pd.DataFrame(columns=['gsm', 'title']) #Create dataframe
    gsm_list, title_list, data_processing_list = [], [], []
    for gsm_name, gsm in gse.gsms.items():
        for key, value in gsm.metadata.items():
            if key == 'title':
                title_list.append(''.join(value))
            if key == 'geo_accession':
                gsm_list.append(''.join(value))
            if key == 'data_processing':
                data_processing_list.append(''.join(value))

    metadata['gsm'], metadata['title'] = gsm_list, title_list
    metadata.columns = range(metadata.shape[1])

    #Output processing style
    print('Data processing summary:\n' + str(set(data_processing_list))) #To determine if all samples have undergone the sample data processing

    #Clean data
    del data.columns.name
    del data.index.name

    #Clean metadata
    metadata[1] = metadata[1].apply(lambda x: x[0:(re.search("\d", x).start()) - 1])

    return data, metadata

"""
DESCRIPTION: Get user file with pertinent sample information
VARIABLES:
file_name= full path of file to import into pandas dataframe
delimiter= delimiter type for importing file, default: ','
axis= Orientiation of the data, where categorical data is either column-wise, (default: 'col') or row-wise ('row'). Case insensitive
sample_ids= Column or row number where sample IDs are found (default: 0)
labels= Column or row number where categorical label data are found (default: 1)
USAGE:
import xpresstools as xp
sample_info = xp.get_info("~/Desktop/sample_info.csv")
ASSUMPTIONS:
Data categories are not labeled
If orientation is not default, it is then specified or else function will not be able to properly format the dataframe for downstream application
"""
def get_info(file_name, delimiter=",", axis="col", sample_ids=0, labels=1):

    #Read in file
    info = pd.read_csv(str(file_name), sep=delimiter, header=None)

    #Reorganize dataframe as necessary so that data is column-wise and
    #sample_ids are the first column, labels are the second column
    if str(axis).lower() == 'col':
        if sample_ids == 0 and labels == 1:
            info = info
        else:
            info = info[[sample_ids, labels]] #Reorder dataframe columns
            info.columns = [0, 1]
    elif str(axis).lower() == 'row':
        info = info.T #Rotate dataframe to make it column-wise
        if sample_ids == 0 and labels == 1:
            info = info
        else:
            info = info[[sample_ids, labels]] #Reorder dataframe columns
            info.columns = [0, 1]
    else:
        print("Incorrect axis option specified")

    return info

"""
DESCRIPTION: Drop samples by sample IDs -- pass in a list of names
VARIABLES:
data= Dataframe containing expression data
ids= List of sample IDs to remove from the dataframe
USAGE:
import xpresstools as xp
df = xp.drop_samples(df, sample_list)
ASSUMPTIONS:
Dataframe axes have been properly formatted (samples are columns, genes are rows)
"""
def drop_samples(data, ids):

    #Check file formats
    if type(ids) is not list:

        return

    #Drop samples in list
    else:
        data_dropped = data.drop(ids, axis=1)

        return data_dropped

"""
DESCRIPTION: Drop samples by label group name
VARIABLES:
data= Dataframe containing expression data
info= Dataframe containing sample information data
label= Name of sample type to drop (string)
USAGE:
import xpresstools as xp
df = xp.drop_label(df, sample_info, "WT")
ASSUMPTIONS:
Dataframe axes have been properly formatted (samples are columns, genes are rows)
Only one string is given to drop per call instance of function
"""
def drop_label(data, info, label):

    #Check file formats
    if type(label) is not str:

        return

    #Drop samples by name (will grab from info df)
    else:
        #Create list of sample_ids based on name provided
        drop_ids = info[info[1] == str(label)]
        drop_ids_list = list(drop_ids[0])

        #Remove these samples from data
        data_dropped = data.drop(drop_ids_list, axis=1)

        return data_dropped

"""
DESCRIPTION: Keep samples by list of label names
VARIABLES:
data= Dataframe containing expression data
info= Dataframe containing sample information data
labels= List of sample types to keep
USAGE:
import xpresstools as xp
df = xp.keep_labels(df, sample_info, ['normal','adenoma'])
ASSUMPTIONS:
Dataframe axes have been properly formatted (samples are columns, genes are rows)
Labels provided are in list format
"""
def keep_labels(data, info, label_list=None):

    if label_list == None:
        label_list = list(set(info[1]))

        keep_ids = info[info[1].isin(label_list)]
        keep_ids_list = list(keep_ids[0])

        #Drop samples not given in list to keep
        data_dropped = data[keep_ids_list]

    else:
        #Check file formats
        if type(label_list) is not list:

            return

        #Keep samples by name (will grab from info df)
        else:

            #Create list of sample_ids based on what is not provided in keep list
            drop_ids = info[~info[1].isin(label_list)]
            drop_ids_list = list(drop_ids[0])

            #Drop samples not given in list to keep
            data_dropped = data.drop(drop_ids_list, axis=1)

    return data_dropped

"""
DESCRIPTION: Rename column names using dictionary

VARIABLES:
data= Dataframe to rename column names
converters= Dataframe where column 0 contains old names and column 1 contains new names
"""
def rename_cols(data, converters):

    data_c = data.copy()
    dictionary = pd.Series(converters[1].values,index=converters[0]).to_dict()
    data_set = data_c.rename(columns=dictionary, inplace=False)
    return data_set

"""
DESCRIPTION: Rename values in a column (selected by providing column name) with a dictionary of keys and sort_values

VARIABLES:
data= Dataframe to rename row values
converters=  Dataframe where column 0 contains old names and column 1 contains new names
label= Name of column to convert names; if 'index' is provided, will rename the index of the dataframe
"""
def rename_rows(data, converters, label='index'):

    data_c = data.copy()

    if label == 'index':
        data_c['index'] = data_c.index
        dictionary = pd.Series(converters[1].values,index=converters[0]).to_dict()
        data_c[label] = data_c['index'].replace(dictionary)
        data_c = data_c.set_index('index')
        del data_c.index.name

    else:
        dictionary = pd.Series(converters[1].values,index=converters[0]).to_dict()
        data_c[label] = data_c[label].replace(dictionary)

    return data_c


"""
DESCRIPTION: Compiles expression counts from multiple files into one table

VARIABLES:
delimiter= Delimiter style for expression files, will also output files if saved in this same format

ASSUMPTIONS:
File length of each is the same and ordered the same (same genes in the same order)
Files to parse are expected to be header-less and column[0] should be gene identifiers and column[1] should be expression values
"""
def catenate_files(directory, file_suffix='txt', save_file=None, delimiter='\t', drop_rows=0):

    #Walk through raw data files within given directory
    if directory[:-1] != '/':
        directory = directory + '/'

    file_list = []
    for subdir, dirs, files in os.walk(directory):
        for f in files:
            if f.endswith(file_suffix): #ignore hidden files and other uninterested files (particular for some submodules)
                file_list.append(f)
            else:
                pass

    #Sort files in alphabetical order (helps in formatting the count tables correctly)
    file_list = sorted(file_list)

    #get gene list from the first, as well as length(#rows)
    with open(str(directory) + str(file_list[0])) as f:
        gene_names = pd.read_csv(f, header=None, usecols=[0], dtype=str, sep=delimiter)
        row_num = len(gene_names) - drop_rows
        if drop_rows > 0:
            gene_names = gene_names[:-drop_rows]

    #populate dataframe with expression values
    data = pd.DataFrame(index=range(row_num))
    data['gene_names'] = gene_names

    for x in file_list:

        with open(str(directory) + x) as f:
            reader = pd.read_csv(f, header=None, usecols=[1], sep=delimiter)
            data[x] = pd.Series(data.index)
            length = len(reader)
            data[x] = reader

    #Remove gene_names label
    data = data.set_index('gene_names')
    del data.index.name

    if save_file != None:
        data.to_csv(str(save_file),sep=delimiter)

    return data

"""
DESCRIPTION: Collate HTseq counts files

VARIABLES:
file_list= List of files with the path names appended to each file to be collated into a single count table
gene_column= Column location in all count files of gene names
gene_column= Column location in all count files of samples
sep= Separator of counts files

ASSUMPTIONS:
No headers are included in the count files
"""
def count_table(file_list, gene_column=0, sample_column=1, sep='\t', drop_rows=5):

    #Read in first count file to get gene names
    df = pd.read_csv(str(file_list[gene_column]), sep=sep, comment='#', header=None)
    pos_starter = [gene_column,sample_column]
    colname = df.columns[pos_starter]
    df = df[colname]

    #For the rest of the files in the file list, add the counts for that sample only
    for f in file_list[1:]:
        df_pull = pd.read_csv(str(f), sep=sep, comment='#', header=None)
        df = pd.concat([df, df_pull[df_pull.columns[sample_column]]], axis=1)
        del df_pull

    #Final formatting clean up of table
    df_counts = df.copy()
    del df
    df_counts = df_counts.set_index(0)
    del df_counts.index.name

    #Remove path and file suffix from each file's name before adding as column names to table
    c = 0
    for x in file_list:
        file_list[c] = x[(x.rfind('/')+1):(x.find('.'))]
        c += 1
    df_counts.columns = file_list
    df_counts = df_counts[:-drop_rows]

    return df_counts



"""
XPRESStools
A toolkit for navigating and analyzing gene expression datasets
alias: xpresstools

Copyright (C) 2019  Jordan A. Berg
jordan <dot> berg <at> biochem <dot> utah <dot> edu

This program is free software: you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

This program is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE. See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along with
this program.  If not, see <https://www.gnu.org/licenses/>.
"""

"""
IMPORT DEPENDENCIES
"""
import csv
import pandas as pd
pd.options.mode.chained_assignment = None
import numpy as np
from multiprocessing import cpu_count, Pool
from functools import partial
from .utils import check_directories

"""
DESCRIPTION: Create a GTF reference file with only protein coding genes and the first n nucleotides of each first exon

RETURNS: Return the truncated, coding only table when running the function

METHODS: Considers strandedness
Multiprocesses chunks of a dataframe on cores of computer

VARIABLES:
input_gtf= GTF reference file path and name
truncate_amount= Number of nucleotides to truncate from the 5' end of each exon 1 (considers strandedness)
save_coding_path= Location to save coding-only GTF, set to None to avoid outputting file
save_truncated_path= Location to save coding-only truncated GTF
sep= Separator type of the GTF file (generally always tab-delimited)

ASSUMPTIONS:
Input file is a properly formatted GTF file
Protein coding transcripts are denoted by 'protein_coding in the final column of the GTF'
"""
def truncate(input_gtf, truncate_amount=45, save_coding_path='./', save_truncated_path='./', sep='\t', return_files=False):

    if save_coding_path != None:
        save_coding_path = check_directories(save_coding_path)
    if save_truncated_path != None:
        save_truncated_path = check_directories(save_truncated_path)

    #Import gtf reference file to
    if str(input_gtf).endswith('.gtf'):
        gtf = pd.read_csv(str(input_gtf), sep=sep, header=None, comment='#', low_memory=False)
    else:
        raise Exception('Error: A GTF-formatted file was not provided')

    #Get only protein_coding coordinates
    gtf_coding = gtf[gtf.iloc[:, 8].str.contains('protein_coding') == True]

    #Save to .gtf file (tsv)
    if save_coding_path != None:
        gtf_coding.to_csv(str(save_coding_path) + 'transcripts_coding.gtf', sep='\t', header=None, index=False, quoting=csv.QUOTE_NONE)

    if truncate_amount != None:
        gtf_coding_c = gtf_coding.copy()

        print("Multiprocessing reference chunks -- this may take a while...")
        gtf_truncated = parallelize_truncator(execute_truncator, gtf_coding_c, truncate_amount)

        if save_truncated_path != None:
            gtf_truncated.to_csv(str(save_truncated_path) + 'transcripts_coding_truncated.gtf', sep='\t', header=None, index=False, quoting=csv.QUOTE_NONE)

    if return_files != False:
        if truncate_amount == None:
            return gtf_coding
        else:
            return gtf_coding, gtf_truncated

"""
DESCRIPTION: GTF truncation function
"""
def execute_truncator(gtf, truncate_amount):

    #gtf[8] = gtf[8].replace({'\"':''}, regex=True)

    gtf['plus'] = gtf[[2,3,4,6,8]].apply(lambda x:
        (x[3] + truncate_amount) if x[2] == "exon" and x[3] + truncate_amount <= x[4] and x[6] == "+" and "exon_number \"1;" in x[8] else (
        "delete_this" if x[2] == "exon" and x[3] + truncate_amount > x[4] and x[6] == "+" and "exon_number \"1;" in x[8] else x[3]),axis=1)

    gtf['minus'] = gtf[[2,3,4,6,8]].apply(lambda x:
        (x[4] - truncate_amount) if x[2] == "exon" and x[3] <= x[4] - truncate_amount and x[6] == "-" and "exon_number \"1;" in x[8] else (
        "delete_this" if x[2] == "exon" and x[3] > x[4] - truncate_amount and x[6] == "-" and "exon_number \"1;" in x[8] else x[4]),axis=1)

    #remove exon1s that are too short
    gtf = gtf[~gtf['plus'].isin(['delete_this'])]
    gtf = gtf[~gtf['minus'].isin(['delete_this'])]

    #copy new coordinates back to original columns
    gtf[3] = gtf['plus']
    gtf[4] = gtf['minus']

    #remove placeholder columns
    gtf = gtf.drop(columns=['plus','minus'])
    return gtf

"""
DESCRIPTION: Parallelization scheme for truncation via chunking dataframe and parallel processing truncation on chunks
"""
def parallelize_truncator(func, *args):

    cores = cpu_count() #Number of CPU cores on your system

    #Split dataframe into chunks for processing by number of cores
    data_split = np.array_split(args[0], cores)
    pool = Pool(cores)

    func = partial(execute_truncator, truncate_amount=args[1])
    data = pd.concat(pool.map(func, data_split))

    pool.close()
    pool.join()

    return data



"""
XPRESStools
A toolkit for navigating and analyzing gene expression datasets
alias: xpresstools

Copyright (C) 2019  Jordan A. Berg
jordan <dot> berg <at> biochem <dot> utah <dot> edu

This program is free software: you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

This program is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE. See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along with
this program.  If not, see <https://www.gnu.org/licenses/>.
"""

"""
IMPORT DEPENDENCIES
"""
import os, sys
import pandas as pd
import numpy as np
import matplotlib
matplotlib.use('Agg')
import matplotlib.pyplot as plt
from sklearn import preprocessing
from .utils_analyze import parallelize, count_threshold_util

"""
INITIALIZATION PARAMETERS
"""
#Retrieve path for scripts used in this pipeline, appended to argument dictionary for every function
__path__, xpresstools_arguments = os.path.split(__file__)

"""
DESCRIPTION: Perform reads per million sample normalization on RNAseq data
"""
def rpm(data):

    data_c = data.copy()
    data_rpm = data_c / (data_c.sum() / 1e6)

    return data_rpm

"""
DESCRIPTION: Perform reads/fragments per kilobase million sample normalization on RNAseq data
"""
def r_fpkm(data, gtf, gene_name_prefix='gene_id \"', gene_name_location=0, sep='\t'):

    #process gtf data for gene_name and gene_length
    gtf = pd.read_csv(str(gtf),sep=sep,comment='#', low_memory=False, header=None)
    gtf_genes = gtf.loc[gtf[2] == 'gene']
    gtf_genes['gene_name'] = gtf[8].str.split(';').str[gene_name_location]
    gtf_genes['length'] = abs((gtf[4]) - (gtf[3]))
    gtf_genes['gene_name'] = gtf_genes['gene_name'].map(lambda x: x.lstrip(gene_name_prefix).rstrip('\"').rstrip(' '))

    #Create dictionary
    length_df = gtf_genes[['gene_name','length']].copy()
    length_df = length_df.set_index('gene_name')
    del length_df.index.name
    length_df = length_df[length_df.index.isin(data.index.values.tolist())]
    length_df.length = length_df.length / 1e3

    #Perform per-million calculations
    data_c = data.copy()
    data_rpm = rpm(data_c)

    #Perform kilobase calculations
    data_rpkm = data_rpm.div(length_df.length, axis=0)
    data_rpkm = data_rpkm.dropna(axis=0)

    return data_rpkm

"""
DESCRIPTION: Perform log2(TE) normalization on ribosome profiling values
ASSUMPTIONS: Table is such that the paired RPF and RNA files are next to each other, in that order
                Values have been properly normalized already (ie. RPM, RPKM)
"""
def te(data, samples=None, log2=True):

    data_c = data.copy()
    data_c += .1

    #Perform translation efficiency calculations
    y = 0
    z = 1
    if samples == None:
        samples = []
        for x in range(int(len(data_c.columns)/2)):
            name = str(data_c.columns[y]) + '_te'
            samples.append(name)
            data_c[name] = data_c[data_c.columns[y]]/data_c[data_c.columns[z]]
            #Move counters for next set of samples
            y = y + 2
            z = z + 2
    else:
        for x in samples:
            data_c[x] = data_c[data_c.columns[y]]/data_c[data_c.columns[z]]
            #Move counters for next set of samples
            y = y + 2
            z = z + 2

    #Get TE_normalized columns
    df_te = data_c[samples]

    #Perform log2 scaling of data
    if log2 == True:
        df_logte = np.log2(df_te)

    return df_logte

"""
DESCRIPTION: Log-scale a sample-normalized dataframe

VARIABLES:
data= Sample normalized, XPRESStools-formated dataframe
log_base= Log-scale to normalize data with (Options: 2 or 10)

ASSUMPTIONS:
Requires a properly formatted dataframe for XPRESStools usage where samples are normalized
"""
def log_scale(data, log_base=10):

    if log_base == 10:
        data_log = np.log10(data + 1)
    elif log_base == 2:
        data_log = np.log2(data + 1)
    else:
        raise Exception('Invalid log_base option provided')

    return data_log

"""
DESCRIPTION: Normalize out batch effects from RNAseq data

VARIABLES:
input_file= Path and file name for sequence dataframe to batch normalize. Output file will be placed in same location and with same prefix, but will include the '_batch.csv' suffix
batch_info= Path and file name for batch effect dataframe. First column should be sample names as in input_file, and second column are the batches. This file should not have a header of any kind
input_sep= Input dataframe delimiter
batch_sep= Batch file delimiter

ASSUMPTIONS:
Data has already been sample normalized
"""
def batch_normalize(input_file, batch_file):

    #Get output file name
    if input_file.endswith('.txt') or input_file.endswith('.tsv'):
        output_file = str(input_file[:-4]) + '_batched.tsv'
    else:
        raise Exception('Unrecognized input_file delimiter type. Files must be tab-delimited')

    #Run sva combat in R
    os.system('rscript ' + str(__path__) + '/batch_normalize.r ' + str(input_file) + ' ' + str(batch_file) + ' ' + str(output_file))

"""
DESCRIPTION: Check sample means and medians
METHODS: Output density plot for dataframe, barplot for each sample
VARIABLES:
USAGE:
ASSUMPTIONS:
Dataframe has been properly formatted so that probes or genes are rows and samples are columns
"""
def check_samples(data):

    wid = len(list(data))
    ax = data.boxplot(column=list(data), figsize=(wid,wid/3))
    ax.set_xlabel('Samples')
    ax.set_ylabel('Expression')

"""
DESCRIPTION: Cleans axis of NULL values
VARIABLES:
USAGE:
ASSUMPTIONS:
If dataframe has been properly formatted previously and genes are in rows, the default parameters will remove along the gene axis
"""
def clean_df(data, axis=0):

    data = data.dropna(axis=axis)
    data = data[~data.index.duplicated(keep=False)]

    return data

"""
DESCRIPTION: Remove genes from analysis where sequence coverage does not meet minimum

VARIABLES:
data= XPRESStools formatted data
minimum= Float or int of minimum count/read value to accept per gene (all samples need to meet this requirement to keep)
"""
def threshold(data, minimum=None, maximum=None):

    data_c = data.copy()
    data_c = parallelize(count_threshold_util, data_c, minimum, maximum)

    return data_c

"""
DESCRIPTION: Prepare dataframes for analysis plotting functions found within analyze.py

METHODS:
Original dataframe is unformatted besides adding labels from info to the first row of the dataframe
Formatted dataframe is scaled if option provided and dataframe is converted to float

VARIABLES:
data= XPRESStools formatted dataframe of expression values
info= XPRESStools formatted sample info dataframe
gene_scale= Scale genes (rows) of data
print_means= Print appropriate means that were scaled for verification

USAGE:
import XPRESStools as mat
df_scaled, df_collapsed = xp.prep_df(df_collapsed, info)

ASSUMPTIONS:
Requires properly formatted df and info dataframes for XPRESStools usage
"""
def prep_data(data, info, gene_scale=True, print_means=False):

    #Convert data to float and drop bad values
    data_c = data.copy()
    data_scaled = data_c.astype(dtype='float')
    data_scaled = clean_df(data_scaled)
    data_c = clean_df(data_c)

    #gene normalization
    if gene_scale == True:
        data_scaled[data_scaled.columns] = preprocessing.scale(data_scaled[data_scaled.columns], axis=1)

    if print_means == True:
        print(data_scaled.mean(axis=1))

    #Map labels to samples
    labels = pd.Series(info[1].values,index=info[0]).to_dict()
    data_c.loc['label'] = data_c.columns.map(labels.get)

    #Output collapsed dataframe
    newIndex = ['label'] + [ind for ind in data_c.index if ind != 'label']
    data_c = data_c.reindex(index=newIndex)

    return data_scaled, data_c



"""
XPRESStools
A toolkit for navigating and analyzing gene expression datasets
alias: xpresstools

Copyright (C) 2019  Jordan A. Berg
jordan <dot> berg <at> biochem <dot> utah <dot> edu

This program is free software: you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

This program is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE. See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along with
this program.  If not, see <https://www.gnu.org/licenses/>.
"""

from setuptools import setup
import re

with open('xpresstools/__init__.py', 'r') as fd:
    version = re.search(r'^__version__\s*=\s*[\'"]([^\'"]*)[\'"]',
                        fd.read(), re.MULTILINE).group(1)

setup(
    name = 'XPRESStools',
    version = version,
    description = 'A toolkit for navigating and analyzing gene expression datasets',
    long_description = open('README.md').read(),
    long_description_content_type='text/markdown',
    author = 'Jordan Berg',
    author_email = 'jordan.berg@biochem.utah.edu',
    url = 'https://github.com/XPRESSyourself/XPRESStools',
    packages = ['xpresstools'],
    exclude= ['tests','docs'],
    package_dir = {'xpresstools': 'xpresstools'},
    license = 'GPL-3.0',
    zip_safe = False,
    install_requires=[
          'pandas',
          'numpy',
          'scipy',
          'seaborn',
          'plotly',
          'matplotlib',
          'sklearn',
          'GEOparse',
          'git-lfs'
      ],

    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Science/Research',
        'Topic :: Scientific/Engineering :: Bio-Informatics'
        ]
)



"""
XPRESStools
A toolkit for navigating and analyzing gene expression datasets
alias: xpresstools

Copyright (C) 2019  Jordan A. Berg
jordan <dot> berg <at> biochem <dot> utah <dot> edu

This program is free software: you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

This program is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE. See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along with
this program.  If not, see <https://www.gnu.org/licenses/>.
"""

"""
IMPORT DEPENDENCIES
"""
import os, sys
import pandas as pd
import numpy as np
from scipy.stats import linregress
from sklearn.decomposition import PCA
import matplotlib
matplotlib.use('Agg')
import matplotlib.pyplot as plt
import seaborn as sns
sns.set(font='arial')
jakes_cmap = sns.diverging_palette(212, 61, s=99, l=77, sep=1, n=16, center='dark') #Custom aesthetics
from .utils_analyze import *
from .interactive import *

"""
INITIALIZATION PARAMETERS
"""
#Retrieve path for scripts used in this pipeline, appended to argument dictionary for every function
__path__, xpresstools_arguments = os.path.split(__file__)

"""
DESCRIPTION: Run Differential Expression analysis using DESeq2
"""
def diff_xpress(input_file, sample_file, equation):

    #Get output file name
    if input_file.endswith('.txt') or input_file.endswith('.tsv'):
        output_file = str(input_file[:-4]) + '_diffx.tsv'
    else:
        raise Exception('Unrecognized input_file delimiter type. Files must be tab-delimited')

    if sample_file.endswith('.txt') or sample_file.endswith('.tsv'):
        pass
    else:
        raise Exception('Unrecognized sample_file delimiter type. Files must be tab-delimited')

    if equation.startswith('~'):
        raise Exception('Tilde should not be included in design formula, script will automatically add this syntax.')

    #Run deseq2 in R
    os.system('rscript ' + str(__path__) + '/diffxpress.r ' + str(input_file) + ' ' + str(sample_file) + ' ' + str(output_file) + ' ' + str(equation))

"""
DESCRIPTION: Plots heatmap for prep_data formatted data

METHODS: We use centroid method as default as it looks at the gene set for each sample holistically to determine a 'centroid' then used for clustering. Other valid options are available (see VARIABLES)

VARIABLES:
data_scaled= Scaled (generally) dataframe as created with the XPRESStools prep_data function
info= XPRESStools formatted info matrix
sample_palette= Dictionary of labels and colors for plotting, or valid seaborns.clustermap col_colors option
gene_info= XPRESStools formatted info matrix for genes (column0) and groupings (column1)
gene_palette= Dictionary of labels and colors for plotting, or valid seaborns.clustermap col_colors option
gene_list= List of genes (either as list or as .csv file path and name with list of genes) to plot (IMPORTANT: Gene names are case-sensitive)
save_fig= If not None, provide full file path, name, and extension to save the file as
dpi= Set dpi of saved figure
bbox_inches= Format saved figure (often useful for making sure no text is cut off)
font_scale= Scaling factor for font
cmap= A valid seaborns.clustermap cmap option (default: Rutter Lab colorblind-friendly scale)
Please see seaborns.heatmap documentation for descriptions of the following options:
center, metric, method, xticklabels, linewidths, linecolor, col_cluster, row_cluster, figsize
figsize= (Width, Height)

USAGE:
import xpresstools as xp
col_colors = {'adenocarcinoma': 'grey',
        'adenoma': (0.8705882352941177, 0.5607843137254902, 0.0196078431372549),
        'normal': (0.00784313725490196, 0.6196078431372549, 0.45098039215686275)}
cbar_kws= {'label': 'z-score',"shrink": 0.2, 'aspect': 10}

xp.heatmap(data_scaled, data_labeled, color_dict=color_dict, gene_list='/path/to/gene_list.csv', figsize=(40,8))

ASSUMPTIONS:
Data has been scaled and labeled with the XPRESStools prep_data function
"""
def heatmap(data, info, sample_palette=None, gene_info=None, gene_palette=None, gene_list=None, col_cluster=True, row_cluster=False, metric='euclidean', method='centroid', font_scale=.8, cmap=jakes_cmap, center=0, xticklabels=True, yticklabels=True, linewidths=0, linecolor='#DCDCDC', cbar_kws=None, figsize=(16,6.5), save_fig=None, dpi=600, bbox_inches='tight'):

    reset_plot(True)
    data_c = analysis_prep(data)

    #Set colors for plotting
    if sample_palette != None:
        sample_palette = prep_palette(info, sample_palette)

    if gene_palette != None and isinstance(gene_info, pd.DataFrame):
        gene_palette = prep_palette(gene_info, gene_palette)

    #Custom panel heatmap
    if gene_list != None:
        data_sub = data_subset(data_c, gene_list)
        plot_data = data_sub.dropna(axis=0)
    else:
        plot_data = data_c.dropna(axis=0)

    #Generate clustermap
    sns.set(font_scale=float(font_scale))
    ax = sns.clustermap(plot_data,
                    cmap=cmap,
                    center=float(center),
                    metric=str(metric),
                    method=str(method),
                    xticklabels=xticklabels,
                    yticklabels=yticklabels,
                    linewidths=float(linewidths),
                    linecolor=linecolor,
                    col_cluster=col_cluster,
                    row_cluster=row_cluster,
                    col_colors=sample_palette,
                    row_colors=gene_palette,
                    figsize=figsize,
                    cbar_kws=cbar_kws
                   )

    #Save figure
    if save_fig is not None:
        plt.savefig(str(save_fig), dpi=dpi, bbox_inches=str(bbox_inches))

"""
DESCRIPTION: Create violin plots of subset of gene expressions or all gene expression by sample

VARIABLES:
data_labeled= Unscaled dataframe with sample labels as created with the XPRESStools prep_data function
gene_list= List of genes (either as list or as .csv file path and name with list of genes) to plot (IMPORTANT: Gene names are case-sensitive)
order= List of samples in order to plot
palette= Dictionary of matplotlib compatible colors for samples
save_fig= If not None, provide full file path, name, and extension to save the file as
dpi= Set dpi of saved figure
bbox_inches= Format saved figure (often useful for making sure no text is cut off)
title= Provide title for figure and saved file if save_fig option used

ASSUMPTIONS:
Data has been scaled and labeled with the XPRESStools prep_data function
"""
def multigene_overview(data, info, palette=None, gene_list=None, order=None, scale='area', save_fig=None, dpi=600, bbox_inches='tight', title=None, grid=False, whitegrid=False):

    reset_plot(whitegrid)
    data_c = data.copy()

    #Custom panel heatmap
    if gene_list != None:
        data_sub = data_subset(data_c, gene_list)
        plot_data = data_sub.dropna(axis=0)
    else:
        plot_data = data_c.dropna(axis=0)

    plot_data_c = analysis_prep(plot_data)
    plot_data_c = label(plot_data_c, info)
    plot_data_c = unstack_data(plot_data_c)

    #Final formatting check
    try:
        plot_data_c[["expr"]] = plot_data_c[["expr"]].apply(pd.to_numeric)
    except:
        raise Exception('Data is not properly formatted for downstream plotting')

    #Plot
    if not order != None and type(order) is list or palette != None and type(palette) is list:
        return

    ax = sns.violinplot(x=plot_data_c['type'], y=plot_data_c['expr'], data=plot_data_c, order=order, palette=palette, scale=scale)
    ax.set_xlabel('')
    ax.set_ylabel('Expression')

    if grid == False:
        ax.grid(False)

    #Save fig
    if save_fig is not None:
        if title is None:
            plt.savefig(str(save_fig), dpi=dpi, bbox_inches=bbox_inches)
        else:
            ax.set_title(str(title))
            plt.savefig(str(save_fig), dpi=dpi, bbox_inches=bbox_inches)

    plt.show()

"""
DESCRIPTION: Plot boxplot overlaid with swarmplot of each sample type's gene expression for the given gene

VARIABLES:
data= Dataframe (can be sample-normalized, prep_data() scaled, or log_scale() scaled)
info= XPRESStools formatted sample info dataframe
gene_name= Name of gene to plot
palette= Dictionary of matplotlib compatible colors for samples
order= List of samples in order to plot
save_fig= If not None, provide full file path, name, and extension to save the file as
dpi= Set dpi of saved figure
bbox_inches= Format saved figure (often useful for making sure no text is cut off)
grid= Control plot gridlines (default: False)

ASSUMPTIONS:
data and info dataframes are properly formatted for XPRESStools and any appropriate sample/gene normalizations have been performed
"""
def gene_overview(data, info, gene_name, palette, order=None, grid=False, whitegrid=False, save_fig=None, dpi=600, bbox_inches='tight'):

    reset_plot(whitegrid)
    data_c = analysis_prep(data).T

    #Prep data_scaled by adding labels from info to column (samples are rows)
    labels = pd.Series(info[1].values,index=info[0]).to_dict()
    data_c['label'] = data_c.index.to_series().map(labels)
    gene_df = data_c[[str(gene_name), 'label']]
    gene_df[[str(gene_name)]] = gene_df[[str(gene_name)]].apply(pd.to_numeric, errors='coerce')

    ax = sns.catplot(x='label', y=str(gene_name), data=gene_df, color='black', order=order, kind='swarm') #Swarm plot
    ax = sns.boxplot(x='label', y=str(gene_name), data=gene_df, width =0.3, fliersize=0, order=order, palette=palette) #Boxplot, fliersize=0 removes outlier diamonds from sns
    plt.setp(ax.collections, sizes=[12]) #Resize markers for catplot

    if grid == False:
        ax.grid(False)

    fig = ax.get_figure()

    if save_fig != None:
        fig.savefig(str(save_fig), dpi=dpi, bbox_inches=bbox_inches)

    plt.show()

"""
DESCRIPTION: Create scatterplot comparing gene expression of two genes across samples

VARIABLES:
data= Dataframe of expression data (samples are columns, genes are rows)
info= XPRESStools formatted sample info dataframe
gene1= Name of gene to plot on x_axis
gene1= Name of gene to plot on y_axis
palette= Dictionary of matplotlib compatible colors for samples
add_linreg= Add linear regression line and print r value in plot title
order_legend= Order of samples to display in legend
title= Plot title to display; if add_linreg used, title variable is void
save_fig= If not None, provide full file path, name, and extension to save the file as
dpi= Set dpi of saved figure
bbox_inches= Format saved figure (often useful for making sure no text is cut off)
grid= Control plot gridlines (default: False)
whitegrid= Use whitegrid background in plot
alpha= Control opacity of points on plot (useful for getting an idea of density in large datasets)

highlight_points= If plotting multiple highlight colors, make sure this and highlight_colors are lists of lists

thresholds can be list of ints or int or None

ASSUMPTIONS:
XPRESStools formatted data and info dataframes, palette is a dictionary of labels and colors to plot points with
"""
def scatter(data, info, x, y, palette=None, add_linreg=False, order_legend=None, title=None, alpha=1, highlight_points=None, highlight_color='DarkRed', highlight_names=None, alpha_highlights=1, size=30, y_threshold=None, x_threshold=None, threshold_color='b', label_points=None, grid=False, whitegrid=False, save_fig=None, dpi=600, bbox_inches='tight', figsize=(10,10)):

    reset_plot(whitegrid)
    plt.figure(figsize=figsize)

    data_c = analysis_prep(data)

    #Prep data_scaled by adding labels from info
    labels = pd.Series(info[1].values,index=info[0]).to_dict()
    data_c.loc['label'] = data_c.columns.map(labels.get)

    if palette == None:
        ax = sns.scatterplot(data_c[str(x)], data_c[str(y)], color='black', alpha=alpha, s=size)
    else:
        #Transpose dataframe so whatever you want to plot is in columns
        if x in data.columns or y in data.columns:
            ax = sns.scatterplot(data_c[str(x)], data_c[str(y)], hue=data_c['label'], palette=palette, alpha=alpha, s=size)
        else:
            ax = sns.scatterplot(data_c.loc[str(x)], data_c.loc[str(y)], hue=data_c.loc['label'], palette=palette, alpha=alpha, s=size)

    if add_linreg == True:
        _x, _y, r_value, title = make_linreg(data_c, x, y)
        ax.plot(_x, _y, '-k')

    #Plot thresholds
    y_threshold = make_threshold_list(y_threshold)
    if type(y_threshold) != list and y_threshold != None or type(y_threshold) == list and any(x is None for x in y_threshold) == False:
        add_y_threshold(y_threshold, threshold_color, ax)

    x_threshold = make_threshold_list(x_threshold)
    if type(x_threshold) != list and x_threshold != None or type(x_threshold) == list and any(x is None for x in x_threshold) == False:
        add_x_threshold(x_threshold, threshold_color, ax)

    #Plot selected genes if user-specified
    if highlight_points != None:
        ax = highlight_markers(data_c, x, y, highlight_points, highlight_color, highlight_names, alpha_highlights, ax)

    #Label points
    if label_points != None and type(label_points) is dict:
        for key, value in label_points.items():
            ax.text(value[0], value[1], str(key), horizontalalignment='left', size='medium', color='black', weight='semibold')

    # Put the legend out of the figure
    ax = make_legend(ax, order_legend, highlight_color, highlight_names)

    plt.xlabel(str(x))
    plt.ylabel(str(y))

    if title != None:
        plt.title(str(title))

    if grid == False:
        plt.grid(False)

    if save_fig != None:
        plt.savefig(str(save_fig), dpi=dpi, bbox_inches=bbox_inches)

    plt.show()

"""
DESCRIPTION: Create scatterplot with r value and jointplot density distributions for axes

VARIABLES:
data= Dataframe of expression data (samples are columns, genes are rows)
info= XPRESStools formatted sample info dataframe
gene1= Name of gene to plot on x_axis
gene1= Name of gene to plot on y_axis
kind= See seaborn jointplot documentation for options (https://seaborn.pydata.org/generated/seaborn.jointplot.html)
palette= Dictionary of matplotlib compatible colors for samples
order= Order of samples to display in plot
save_fig= If not None, provide full file path, name, and extension to save the file as
dpi= Set dpi of saved figure
bbox_inches= Format saved figure (often useful for making sure no text is cut off)
grid= Control plot gridlines (default: False)
whitegrid= Use whitegrid background in plot

Full functionality only supported for kind='reg' currently

ASSUMPTIONS:
XPRESStools formatted data and info dataframes, palette (if used) is a dictionary of labels and colors to plot points with
"""
def jointplot(data, info, x, y, kind='reg', palette=None, order=None, title_pad=0, title_pos='right', grid=False, whitegrid=False, save_fig=None, dpi=600, bbox_inches='tight', figsize=(10,10)):

    reset_plot(whitegrid)
    plt.figure(figsize=figsize)

    data_c = data.copy()
    data_c = analysis_prep(data_c)

    #Prep data_scaled by adding labels from info
    if 'label' not in data_c.index:
        labels = pd.Series(info[1].values,index=info[0]).to_dict()
        data_c.loc['label'] = data_c.columns.map(labels.get).T

    #Get r
    gene_a = get_array(data_c, x)
    gene_b = get_array(data_c, y)

    #Plot
    ax = sns.jointplot(x=gene_a, y=gene_b, kind=kind)
    ax.ax_joint.collections[0].set_visible(False)

    if kind.lower() == 'reg':
        r_value = stats.pearsonr(gene_a, gene_b)[0]

        ax = sns.scatterplot(x=str(x), y=str(y), data=data_c.T, hue='label', palette=palette, hue_order=order)

        plt.xlabel(str(x))
        plt.ylabel(str(y))
        ax.set_title('r: ' + str(round(r_value,2)), pad=float(title_pad), loc=title_pos.lower())

    if grid == False:
        plt.grid(False)

    if kind.lower() == 'reg':
        fig = ax.get_figure()
    plt.show()

    if save_fig != None:
        plt.savefig(str(save_fig), dpi=dpi, bbox_inches=bbox_inches)

"""
DESCRIPTION: Calculates r, r^2 values, and p-values for every gene against target gene for given dataset

VARIABLES:
data= Dataframe of expression data (samples are columns, genes are rows)
gene_name= Gene to run correlations against
save_file= File path and name to save correlations matrix with
delimiter= Delimiter to use for save_file

ASSUMPTIONS:
data dataframes is properly formatted for XPRESStools, sample scaling required but gene scaling is not necessary
data should not contain labels
"""
def linreg(data, gene_name, save_file, delimiter=','):

    data_c = data.copy()
    data_c = analysis_prep(data_c)

    if 'label' in data_c.index:
        data_c = data_c.drop(labels='label',axis=0)

    #Gene expression array for gene of interest
    interest = get_array(data_c, gene_name)

    #Get expression arrays for all genes and run linear regression, save values to matrix
    lm_interest=[]
    for row in data_c.iterrows():
        index, data = row
        gene = get_array(data_c, index)

        if len(gene) is not len(interest):
            continue
        else:
            slope, intercept, r_value, p_value, std_err = linregress(gene, interest)
            lm_interest.append([index, slope, intercept, r_value, (r_value**2), p_value, std_err])

    df_lm_interest = pd.DataFrame(lm_interest, columns=['gene', 'slope', 'intercept', 'r_value', 'r_squared', 'p_value', 'std_err'])

    # Save linear modeling metrics to .csv
    df_lm_interest.to_csv(str(save_file), sep=delimiter)

"""
DESCRIPTION: Plot volcano plot for dataframe, can highlight subset of genes

VARIABLES:
data= Sample-normalized, XPRESStools-formatted data -- should NOT be gene-normalized
info= XPRESStools formatted sample info dataframe
label_comp= Sample label to compare against base
label_base= Sample label to use as base for comparison
highlight_points= If provided with a list, or a file path and name to a .csv-type series of gene names, will highlight genes of interest in a different color (Gene names are case-sensitive). Must be a list of lists to allow for multiple highlighting groups
highlight_color= Color to use for highlighted genes
y_threshold= -log10(P-Value) threshold to use to identify significant hits. Will create a dotted line on the plot and use to pull significant hits if return_threshold_hits is not None
x_threshold= log2(Fold Change) threshold to use to identify significant hits. Will take the positive and negative of the number. Will create a dotted line on the plot and use to pull significant hits if return_threshold_hits is not None
save_threshold_hits= Will save a .csv-type matrix of significant genes/hits to the file path and name specified
save_threshold_hits_delimiter= Delimiter to use in exporting return_threshold_hits
save_fig= If not None, provide full file path, name, and extension to save the file as
dpi= Set dpi of saved figure
bbox_inches= Format saved figure (often useful for making sure no text is cut off)
return_data= Return dataframe with PCs (will not print plot to stout)
label_genes= Dictionary for add labels to non-plotly volcano plot

Need to output threshold to figure out coordinates to label

LIMITATIONS:
Will only perform comparison between 2 sample types

ASSUMPTIONS:
data should ONLY be sample normalized. If using a previous function that returned a modified original
y_threshold must be a postive integer or float
"""
def volcano(data, info, label_comp, label_base, order_legend=None, title=None, alpha=1, highlight_points=None, highlight_color='DarkRed', highlight_names=None, alpha_highlights=1, size=30, y_threshold=None, x_threshold=None, threshold_color='b', save_threshold_hits=None, save_threshold_hits_delimiter=',', label_points=None, grid=False, whitegrid=False, return_data=False, plotly_login=False, save_fig=None, dpi=600, bbox_inches='tight', figsize=(10,10)):

    reset_plot(whitegrid)

    data_c = analysis_prep(data)
    info_c = info.copy()

    #Add labels to column name
    if 'label' in data_c.index:
        data_c = data_c.drop(labels='label',axis=0)

    #Convert sample names to include their category (ie: Sample1_Normal, Sample2_Treatment)
    info_c["id"] = info_c[0] + '_' + info_c[1]
    labels = pd.Series(info_c['id'].values,index=info_c[0]).to_dict()
    data_c = data_c.rename(labels, axis='columns')

    data_c += 1e-7 #Avoid divide by 0 error

    #Calculate parameters
    data_c = parallelize(calculate_fc, data_c, label_comp, label_base)
    data_c = parallelize(calculate_p, data_c, label_comp, label_base)

    #Plot all genes
    if plotly_login == False:
        scatter(data_c, info_c, 'log$_2$(Fold Change)', '-log$_1$$_0$(P-Value)', palette=None, add_linreg=False, order_legend=order_legend, title=title, save_fig=save_fig, dpi=dpi, bbox_inches=bbox_inches, grid=grid, whitegrid=whitegrid, alpha=alpha, highlight_points=highlight_points, highlight_color=highlight_color, highlight_names=highlight_names,  alpha_highlights=alpha_highlights, y_threshold=y_threshold, x_threshold=x_threshold, threshold_color=threshold_color, label_points=label_points, size=size, figsize=figsize)

    else:
        data_c = data_c.rename(columns = {'log$_2$(Fold Change)':'log2 Fold Change', '-log$_1$$_0$(P-Value)':'-log10 P-Value'})
        interactive_scatter(data_c, info_c, x='log2 Fold Change', y='-log10 P-Value', plotly_login=plotly_login, file_name=save_fig, highlight='sample', palette=None)

    #Output threshold hits outer bounds, specific to volcano plot
    if save_threshold_hits != None:
        output_threshold(data_c, x_threshold, y_threshold, save_threshold_hits, save_threshold_hits_delimiter)

    #Save hits if user-specified
    if return_data == True:
        df_c = data_c[['log2 Fold Change', '-log10 P-Value']].copy()
        return df_c

"""
DESCRIPTION: Plot a 2-D PCA with confidence intervals or a 3-D PCA with no confidence intervals

RETURNS: Dataframe with PCs calculated

METHODS: 3-D PCA -- option to output as interactive plot by providing plotly credentials, or as a static plot without these credentials (default)

VARIABLES:
data_scaled= Scaled dataframe as created with the XPRESStools prep_data and scaling function -- can be a dataframe prepared using the prep_data() function or log_scale() function
info= XPRESStools formatted sample info dataframe
palette= Dictionary of matplotlib compatible colors for samples; for plotly 3-D PCA,
grouping= Perform PCA sample-wise (default) or gene-wise (grouping='genes')
gene_list= List of genes (either as list or as .csv file path and name with list of genes) to plot (IMPORTANT: Gene names are case-sensitive) (only functional when grouping='samples')
gene_labels= Option for grouping='genes', not currently implemented
ci= For 2-D PCA, confidence interval to display for each sample type (i.e. 1 == CI1 == 68%, 2 == CI2 == 95%, 3 == CI3 == 99%)
principle_components= Principle components to plot (default: [1,2] for 2-D PCA, or [1,2,3] for 3-D PCA)
n_components= Number of components to calculate in dataframe (more applicable if you want to perform a deeper survey of the principle components)
_3d_pca= Plot 3 principle components (default: False)
plotly_login= ['userid','api key'], usage of this option creates a plotly interactive plot that can be viewed online using your login credentials
title= Provide title for figure and saved file if save_fig option used
save_fig= If not None, provide full file path, name, and extension to save the file as. if saving scree using this, don't need extension
dpi= Set dpi of saved figure
bbox_inches= Format saved figure (often useful for making sure no text is cut off)
order_legend= List of integers to reorder samples in legend (i.e. if samples are displayed 1:Sample_A, 2:Sample_C, 3:Sample_B, provide the list [1,3,2]) (Not currently compatible with 3-D PCA options)
grid= For non-plotly options, remove gridlines from plot
fig_size= Option not used in function currently
size= Marker size

palette examples:
colors = {'adenocarcinoma': (0.5725490196078431, 0.5843137254901961, 0.5686274509803921),
        'adenoma': (0.8705882352941177, 0.5607843137254902, 0.0196078431372549),
        'normal': (0.00784313725490196, 0.6196078431372549, 0.45098039215686275)}
colors = {'adenocarcinoma': 'gray',
        'adenoma': 'orange',
        'normal': 'green'}
colors = {'adenocarcinoma': '#59656d',
        'adenoma': '#f0833a',
        'normal': '#0a5f38'}

USAGE:


ASSUMPTIONS:
Plotly and api key generation, addition
Data has been scaled and labeled with the XPRESStools prep_data function

FEATURES TO ADD:
Allow for compatibility with adding labels for gene classes for plotting
Option to order legend and rearrange in 3d static plot
Add options to vary marker size and opacity
"""
def pca(data, info, palette, grouping='samples', gene_list=None, gene_labels=False, _3d_pca=False, principle_components=[1,2], n_components=10, ci=2, scree_only=False, save_scree=False, size=30, order_legend=None, title=None, figsize=(10,10), grid=False, whitegrid=False, save_fig=None, dpi=600, bbox_inches='tight', return_pca=False, plotly_login=None):

    highlight_color = None
    highlight_names = None

    reset_plot(whitegrid)
    principle_components = init_pca(principle_components, _3d_pca, plotly_login)

    data_c = data.copy()
    data_c = analysis_prep(data)

    #Add labels to column name
    if 'label' in data_c.index:
        data_c = data_c.drop(labels='label',axis=0)

    #Custom panel heatmap
    if gene_list != None:
        data_sub = data_subset(data_c, gene_list)
        plot_data = data_sub.dropna(axis=0)
    else:
        plot_data = data_c.dropna(axis=0)

    #Format data for sample-wise or gene-wise PCA analysis
    if str(grouping) == 'samples':
        plot_data = plot_data.T
    elif str(grouping) == 'genes':
        plot_data = plot_data
        print('PCA by genes is not yet implemented...')
        return
    else:
        print('Incorrect grouping variable provided')
        return

    #Prep PCA
    pca = PCA(n_components=n_components)
    pca_result = pca.fit_transform(plot_data.values)

    #Record PCs
    x = 1
    while x <= n_components:
        pc = 'PC' + str(x)
        component = x - 1
        plot_data[pc] = pca_result[:,component]
        x += 1

    #Scree
    if save_scree == True or scree_only == True:
        scree = make_scree(pca, n_components, save_fig, dpi, bbox_inches, scree_only, save_scree, grid, whitegrid)
        if scree_only == True:
            return
    else:
        scree = np.round(pca.explained_variance_ratio_, decimals=4)*100

    if gene_labels == False:

        #Prep data_scaled by adding labels from info to column (samples are rows)
        labels = pd.Series(info[1].values,index=info[0]).to_dict()
        plot_data['label'] = plot_data.index.to_series().map(labels)

        #Plot PCa & PCb
        pc_list = []
        for p in principle_components:
            pc_list.append('PC' + str(p))

        pc_list.append('label')

        df_pca = plot_data[pc_list] #Prepare pca data

        #2D PCA
        if _3d_pca == False:
            df_pca.columns = ['PCa', 'PCb', 'label']
            unique_labels = df_pca['label'].unique() #Gather unique labels

            #Non-interactive
            if plotly_login == None:
                pca2(df_pca, unique_labels, palette, principle_components, scree, order_legend, save_fig, dpi, bbox_inches, ci, grid, title, size, highlight_color, highlight_names)

            #Plotly
            else:
                interactive_scatter(df_pca, info, 'PCa', 'PCb', plotly_login, palette, highlight='sample', save_fig=save_fig)

        #3D PCA
        elif _3d_pca == True:

            #Plotly
            if plotly_login != None:
                interactive_3D(df_pca, size, palette, plotly_login, save_fig)

            #Non-interactive
            else:
                pca3(df_pca, palette, save_fig, dpi, bbox_inches, size, order_legend)

        else:
            return

    #Gene scatter
    else:
        print('This feature has not been implemented yet')

    if return_pca == True:
        return df_pca
